#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
This module defines the interfaces that the rest of the redesign uses.

1. LocationReport constitutes all the information about the products we will be fetching
2. LocatedFile is an entry in the LocationReport for a specific file
3. FileFetcher is the interface for fetching a single file
4. FileValidator is the interface for validating a file to ensure it was fetched correctly
5. FetcherFactory is a bridge between the LocationReport and a sequence of FileFetchers,
   which feeds the next steps
6. FetchPlan aggregates FileFetchers together

The plan of operation is expressed by these interfaces like so:

1. A LocationReport is produced
2. The FetcherFactory is given to the LocationReport, to create a series of FileFetchers
3. The FileFetchers are aggregated into a FetchPlan
4. The FetchPlan is executed

"""
import abc
from abc import ABC, abstractmethod
from enum import Enum
from pathlib import Path
from typing import List, NamedTuple, Union, Optional, Dict

from .exceptions import NormalModeFileExistsError
from .retry import retry


# pylint: disable=C0116, E0239, E1101, R0903, W0107


class ForceMode(Enum):
    """
    ForceMode is either normal or force. In normal mode, we refuse to overwrite an
    existing file. In force mode, we are more than willing to destroy your old data.
    """

    NORMAL = False
    FORCE = True

    def check_destination(self, path: Path):
        if self == self.NORMAL and path.exists():
            raise NormalModeFileExistsError(f"Cannot fetch to {path} because the file already exists.")


class LocatedFile(abc.ABC):
    """
    Located files know some information about themselves, which gives them two important capabilities:

    1. Knowing how to construct a FileFetcher for them, given a FetcherFactory, and
    2. Knowing whether or not they share resources with some other LocatedFile.

    The exact details of the LocatedFile will vary as the location report changes over time, but
    the canonical implementations right now are OracleXml and NgasFile.
    """

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """
        The name of the file
        :return:
        """
        pass

    @property
    @abc.abstractmethod
    def size(self) -> int:
        """
        The size of the file
        :return:
        """
        pass

    @abc.abstractmethod
    def make_fetcher(self, factory: "FetcherFactory") -> "FileFetcher":
        """
        Convert this file into the appropriate kind of fetcher, with the assistance of the fetcher factory.
        :param factory:
        :return:
        """
        pass

    @abc.abstractmethod
    def shares_resources_with(self, other: "LocatedFile") -> bool:
        """
        True if these two files depend on the same underlying servers (and thus should not be performed in parallel).

        :param other: the other file we are comparing with
        :return:
        """
        pass

    @property
    @abc.abstractmethod
    def destination(self) -> Path:
        """
        Most LocatedFiles have a subdirectory and relative_path and this is how you piece them together
        to form the delivery destination.
        :return:
        """

    @property
    @abc.abstractmethod
    def science_product_locator(self) -> str:
        """
        The science product locator for this file, if we happen to know it.
        :return:
        """
        pass

    @property
    @abc.abstractmethod
    def is_sdm(self) -> bool:
        """
        True if the file under discussion is an SDM (for SDM-only mode)
        """
        pass


class FileFetcher(ABC):
    """
    A FileFetcher fetches a single file from some external system.
    FileFetchers can have Validators attached to them, which will
    be called in order after the fetch completes to validate that
    the fetch was successful.

    The FetchPlan system relies on FileFetchers knowing when they
    required shared access to the external resource. FileFetchers
    implement a method to indicate whether they share resources,
    and if they do, FetchPlan groups them together to limit their
    concurrency. FetchPlan assumes that FileFetchers are disjoint
    in their use of shared resources; i.e. that if you have two
    fetchers that do not share resources with each other, there
    will never be a third fetcher that shares resources with both.
    """

    def __init__(self):
        self.validators = []

    @property
    @abstractmethod
    def file(self) -> LocatedFile:
        """
        Return the located file used by this fetcher

        :return: LocatedFile used by this fetcher
        """
        pass

    def destination(self, context: "FetchContext") -> Path:
        """
        Return the actual destination of this file, using the context.

        :param context: the context in which we are fetching.
        :return:
        """
        return context.path_to(self.file)

    @abstractmethod
    def do_fetch(self, context: "FetchContext") -> Path:
        """
        Fetches something.

        :return: result
        :raise RetryableFetchError: if something fails that can be retried
        :raise FetchError: in other cases when something goes sideways
        """
        pass

    @abstractmethod
    def shares_resources_with(self, other: "FileFetcher") -> bool:
        """
        True if this and the other FileFetcher access the same shared resources
        (such as a web host or database).

        FetchPlan assumes that resource sharing is a transitive property,
        i.e., that if fetcher A and fetcher B share resources, and so do
        fetchers A and C, then fetchers B and C share resources as well.

        :param other: another FileFetcher
        :return:  true if these two use the same shared resources
        """
        pass

    def and_validate_with(self, validator: "FileValidator") -> "FileFetcher":
        """
        Attach a validator to this FileFetcher. Can be called multiple times to attach many validators.

        :param validator:  validator to attach
        :return: self, so that calls can be chained
        """
        self.validators.append(validator)
        return self

    def fetch(self, context: "FetchContext"):
        """
        Fetch a file and validate it.

        :return: path to the newly fetched file, if successful
        :raise FileValidationFault: if the fetch did not succeed
        """
        # do the fetch
        result = self.do_fetch(context)

        # now validate with our validators, if we fetched a file
        if result is not None:
            for validator in self.validators:
                retry(lambda: validator.validate(result), [1, 4, 9])

        # if we made it here, we have a result to report
        context.fetch_reporter.complete(self.file, result)

        # and return
        return result


class FetcherFactory(ABC):
    """
    As the name suggests, a FetcherFactory is a device for constructing FileFetchers.
    The LocationReport is able to use this to produce a list of FileFetchers of various types,
    possibly according to settings provided by the user at the command-line.
    """

    @abstractmethod
    def fetch_ngas(self, file: LocatedFile) -> FileFetcher:
        """
        Fetch a single file from NGAS

        :param file:  the file to fetch
        :return: a fetch plan to retrieve that file
        """
        pass

    @abstractmethod
    def fetch_oracle_xml(self, file: LocatedFile) -> FileFetcher:
        """
        Fetch a single ALMA SDM from their Oracle database.

        :param file: SDM as XML from ALMA DB
        :return: an aggregating FetchPlan
        """
        pass

    @abstractmethod
    def fetch_plan(self, fetchers: List[Union[FileFetcher, "FetchPlan"]], concurrency: int = 1) -> "FetchPlan":
        """
        Run a number of other plans or fetches in a larger FetchPlan.

        :param fetchers:  fetchers or plans
        :param concurrency:  the number to run concurrently
        :return: a new FetchPlan
        """
        pass


class FileValidator(ABC):
    """
    A FileValidator examines a file on disk and verifies that it passes muster
    in some implementation-dependent manner.
    """

    @abstractmethod
    def validate(self, path: Path):
        """
        Validates a file on the filesystem. If the file is OK, the validator will simply return;
        if the file is not OK an exception will be raised.

        :raise FileValidationFault: if the file fails to validate
        """
        pass


class LocationReport(NamedTuple):
    """Represents a locations report from archiveService"""

    files: List[LocatedFile]
    aggregate_size: int
    ancillaries: Dict[str, "LocationReport"] = []

    def fetchers(self, file_enumerator: "LocationReportFileEnumerator", factory: FetcherFactory) -> List[FileFetcher]:
        return [file.make_fetcher(factory) for file in file_enumerator.files(self)]


class LocationReportFileEnumerator(ABC):
    """Obtains files from a report"""

    @abstractmethod
    def files(self, location_report: LocationReport) -> List[LocatedFile]:
        """
        Obtain the list of files from this location report
        :param location_report: the location report to examine
        :return: a list of LocatedFiles
        """
        pass


class SizeCalculator(ABC):
    """Obtains the total size in bytes of things to fetch from a report"""

    @abstractmethod
    def aggregate_size(self, report: LocationReport) -> int:
        """
        Obtain the total sizes of things that are going to be fetched from this report.
        :param report: report to examine
        :return: the size, in bytes
        """
        pass


class FetchContext(ABC):
    """
    Represents the context in which we are fetching.
    """

    @property
    @abc.abstractmethod
    def destination(self) -> Path:
        """
        The directory into which we should be writing our products
        """

    @property
    @abc.abstractmethod
    def force(self) -> ForceMode:
        """
        Whether or not to overwrite files in the destination directory.
        :return:
        """

    @property
    @abc.abstractmethod
    def concurrency(self) -> int:
        """
        The amount of concurrency we should support.
        :return: number of threads we can use
        """

    @property
    @abc.abstractmethod
    def fetch_reporter(self) -> "FetchProgressReporter":
        """
        Object to report progress with.
        :return: progress reporter
        """

    @abc.abstractmethod
    def locator_fetch_destination(self, product_locator: Optional[str]) -> Path:
        """
        Returns the destination for products related to this product locator
        :param product_locator:  the product locator to find the destination for
        :return:
        """
        pass

    def path_to(self, file: LocatedFile) -> Path:
        """
        Path to the file. The LocatedFile knows where it wants to be placed, but does not know where we are
        fetching to. This combines the two pieces of information.

        :param file: file to ascertain the path to
        :return: path for this file in the destination area
        """
        return self.destination / file.destination

    def check_and_prepare_destination(self, file: LocatedFile) -> Path:
        """
        Ensure that the destination for this file is acceptable. Generate missing subdirectories.

        :param file: file to check
        :return: Path to the file we are expecting
        :raises NormalModeFileExistsError: if the destination is unacceptable for this mode
        """
        # compute the final destination for the file
        destination = self.path_to(file)

        # check it, using the force mode
        self.force.check_destination(destination)

        # generate the parent directories as necessary
        destination.parent.mkdir(parents=True, exist_ok=True)
        destination.parent.chmod(0o770)

        return destination


class Locator(ABC):
    """Obtains a location report in various ways"""

    @abstractmethod
    def locate(self) -> LocationReport:
        """
        Obtain the LocationReport we need to proceed.

        :return: a LocationReport for where all the files we need are
        """


class ProductMetadata(NamedTuple):
    telescope: str
    project_code: str
    type: str
    segment: str
    mous_id: str
    date: float

    def __json__(self):
        return dict(
            telescope=self.telescope,
            project_code=self.project_code,
            type=self.type,
            segment=self.segment,
            mous_id=self.mous_id,
            date=self.date,
        )


class ProductMetadataFinder(ABC):
    """Obtains product metadata for a given product locator"""

    @abstractmethod
    def retrieve_metadata(self, product_locator: str) -> ProductMetadata:
        """
        Obtain the product metadata for this locator
        :param product_locator:
        :return:
        """


class FetchProgressReporter(ABC):
    """
    A progress reporter receives updates from the fetchers, which they can use to
    display some kind of useful output.
    """

    def __init__(self, expected_size: int):
        pass

    @abstractmethod
    def complete(self, file: LocatedFile, path: Path):
        pass


class FetchPlan(ABC):
    """
    A fetch plan is a lot like a FileFetcher, except it potentially does many files
    and thus we have no specific return value from the fetch method.
    """

    @abstractmethod
    def fetch(self, context: FetchContext):
        pass


class FetchPlanDecorator(FetchPlan):
    def __init__(self, underlying: FetchPlan):
        self.underlying = underlying

    def fetch(self, context: FetchContext):
        self.underlying.fetch(context)
